use std::{collections::HashMap, fs, path::Path, str::FromStr};

use fluent_bundle::{concurrent::FluentBundle, FluentArgs, FluentResource, FluentValue};
use unic_langid::LanguageIdentifier;

#[derive(Clone)]
pub struct Localization {
    bundles: std::sync::Arc<HashMap<LanguageIdentifier, FluentBundle<FluentResource>>>,
}

impl Localization {
    pub fn init(locale_directory: &Path) -> Self {
        let mut bundles = HashMap::new();
        let directory_content =fs::read_dir(locale_directory)
            .expect("Unable to read locales directory");

        for lang_dir in directory_content {
            let lang_dir = lang_dir.expect("I/O error");
            if lang_dir.path().is_dir() {
                let langid: LanguageIdentifier = lang_dir.file_name()
                    .into_string()
                    .expect("String convertion failed")
                    .parse()
                    .expect("Failed to parse language tag");

                let mut bundle = FluentBundle::new_concurrent(vec![langid.clone()]);

                let directory_content = fs::read_dir(lang_dir.path())
                    .expect("Unable to read locales directory");
                for resource in directory_content {
                    let resource = resource.expect("I/O error");

                    if resource.path().is_file() {
                        let resource = fs::read_to_string(resource.path())
                            .expect("Failed to open file");

                        let resource = FluentResource::try_new(resource)
                            .expect("Failed to parse an FTL string.");

                        bundle
                            .add_resource(resource)
                            .expect("Failed to add FTL resources to the bundle.");
                    }

                }

                bundles.insert(langid, bundle);
            }
        }

        Localization {
            bundles: std::sync::Arc::new(bundles),
        }
    }

    pub fn language_middleware(&self, default_locale: LanguageIdentifier) -> ExtractLanguageLayer {
        ExtractLanguageLayer {
            default_locale,
            supported_locales: std::sync::Arc::new(self.bundles.keys().cloned().collect())
        }
    }
}

impl tera::Function for Localization {
    fn call(&self, args: &HashMap<String, tera::Value>) -> tera::Result<tera::Value> {
        let locale = args.get("lang")
            .and_then(|lang| lang.as_str())
            .and_then(|lang| LanguageIdentifier::from_str(lang).ok())
            .ok_or(tera::Error::msg("localize: Missing lang parameter"))?;

        let msg = args.get("msg")
            .and_then(|lang| lang.as_str())
            .ok_or(tera::Error::msg("localize: Missing msg parameter"))?;

        let attribute = args.get("attr")
            .and_then(|lang| lang.as_str());

        let bundle = self.bundles.get(&locale)
            .ok_or_else(|| tera::Error::msg(format!("localize: Could not find locale {locale}")))?;

        let message = bundle.get_message(msg)
            .ok_or_else(|| tera::Error::msg(format!("localize: Could not find message {msg}")))?;


        let pattern = if let Some(attribute) = attribute {
            message.get_attribute(attribute)
                .map(|attribute| attribute.value())
                .ok_or_else(|| tera::Error::msg(format!("localize: Attribute {msg}.{attribute} has no value")))?
        } else {
            message.value()
                .ok_or_else(|| tera::Error::msg(format!("localize: Message {msg} has no value")))?
        };

        let mut msg_args = FluentArgs::new();

        for (key, value) in args {
            if key != "msg" && key != "lang" && key != "attr" {
                msg_args.set(key, fluent_value_from_tera(value, key)?);
            }
        }

        let mut errors = Vec::new();

        let localized_message = bundle.format_pattern(pattern, Some(&msg_args), &mut errors);

        for err in errors {
            eprint!("[warn] localization error: {err}");
        }

        Ok(tera::Value::from(localized_message))

    }
}

fn fluent_value_from_tera<'s>(value: &'s tera::Value, name: &str) -> tera::Result<FluentValue<'s>> {
    match value {
        tera::Value::String(string) => Ok(FluentValue::from(string)),
        tera::Value::Number(number) => Ok(FluentValue::from(number.as_f64())),
        _ => Err(tera::Error::msg(format!("localize: Argument {name} can only be a string or a number"))),
    }
}

#[derive(Clone)]
pub struct ExtractLanguageLayer {
    default_locale: LanguageIdentifier,
    supported_locales: std::sync::Arc<Vec<LanguageIdentifier>>,
}

impl<S> tower::Layer<S> for ExtractLanguageLayer {

    type Service = ExtractLanguageService<S>;

    fn layer(&self, inner: S) -> Self::Service {
        ExtractLanguageService {
            inner,
            default_locale: self.default_locale.clone(),
            supported_locales: self.supported_locales.clone(),
        }
    }
}

#[derive(Clone)]
pub struct ExtractLanguageService<S> {
    inner: S,
    default_locale: LanguageIdentifier,
    supported_locales: std::sync::Arc<Vec<LanguageIdentifier>>,
}

impl<S> ExtractLanguageService<S> {
    // https://httpwg.org/specs/rfc9110.html#field.accept-language
    pub fn language_from_header(&self, headers: &axum::http::HeaderMap) -> LanguageIdentifier {
        let lang_preferences = headers
            .get("Accept-Language")
            .and_then(|val| val.to_str().ok());

        if let Some(lang_preferences) = lang_preferences {
            let mut languages = Vec::new();

            for lang_item in lang_preferences.split(",") {
                let lang_item = lang_item.trim();
                let lang_config = lang_item.split_once(';');
                if let Some((lang_id, config))  = lang_config {
                    let lang_id = lang_id.trim();
                    let preference = config.trim().strip_prefix("q=")
                        .and_then(|value| value.parse::<f32>().ok());
                    if let Some(preference) = preference {
                        languages.push((lang_id.trim(), preference));
                    }
                } else {
                    languages.push((lang_item, 1f32));
                }
            }

            languages.sort_by(|l1, l2| l1.1.total_cmp(&l2.1));

            let iter = languages.into_iter()
                .rev()
                .filter(|l| l.1 >= 0.001 && l.1 <= 1f32)
                .filter_map(|(lang, _)| LanguageIdentifier::from_str(lang).ok());

            for lang in iter {
                for supported_lang in self.supported_locales.iter() {
                    if lang == *supported_lang {
                        return lang;
                    }
                }
            }

            self.default_locale.clone()
        } else {
            self.default_locale.clone()
        }
    }
}

impl<S> tower::Service<axum::extract::Request> for ExtractLanguageService<S>
where
    S: tower::Service<axum::extract::Request, Response = axum::response::Response> + Send + 'static,
    S::Future: Send + 'static,
{
    type Response = S::Response;
    type Error = S::Error;
    // `BoxFuture` is a type alias for `Pin<Box<dyn Future + Send + 'a>>`
    type Future = std::pin::Pin<Box<dyn Future<Output = Result<Self::Response, Self::Error>> + Send + 'static>>;

    fn poll_ready(&mut self, cx: &mut std::task::Context<'_>) -> std::task::Poll<Result<(), Self::Error>> {
        self.inner.poll_ready(cx)
    }

    fn call(&mut self, mut req: axum::extract::Request) -> Self::Future {
        let headers = req.headers();

        let lang = self.language_from_header(headers);

        req.extensions_mut().insert(lang);

        Box::pin(self.inner.call(req))
    }
}
