use std::{collections::HashMap, fmt};

use serde::{Deserialize, Serialize, Serializer, ser::SerializeStruct};

use crate::resources::dns::internal;


#[derive(Debug, Deserialize, Serialize, PartialEq)]
#[serde(rename_all="lowercase")]
pub enum FriendlyRType {
    Address,
    Alias,
    MailServer,
    NameServer,
    Service,
    Spf,
    TextData,
}

impl fmt::Display for FriendlyRType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            FriendlyRType::Address => write!(f, "address"),
            FriendlyRType::Alias => write!(f, "alias"),
            FriendlyRType::MailServer => write!(f, "mailserver"),
            FriendlyRType::NameServer => write!(f, "nameserver"),
            FriendlyRType::Service => write!(f, "service"),
            FriendlyRType::Spf => write!(f, "spf"),
            FriendlyRType::TextData => write!(f, "textdata"),
        }
    }
}

pub trait HasRType {
    fn rtype(&self) -> FriendlyRType;
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(tag = "_type")]
pub enum FriendlyRData {
    Address(Address),
    Alias(Alias),
    MailServer(MailServer),
    NameServer(NameServer),
    Service(Service),
    Spf(Spf),
    TextData(TextData),
}

impl HasRType for FriendlyRData {
    fn rtype(&self) -> FriendlyRType {
        match self {
            FriendlyRData::Address(_) => FriendlyRType::Address,
            FriendlyRData::Alias(_) => FriendlyRType::Alias,
            FriendlyRData::MailServer(_) => FriendlyRType::MailServer,
            FriendlyRData::NameServer(_) => FriendlyRType::NameServer,
            FriendlyRData::Service(_) => FriendlyRType::Service,
            FriendlyRData::Spf(_) => FriendlyRType::Spf,
            FriendlyRData::TextData(_) => FriendlyRType::TextData,
        }
    }
}

#[derive(Debug)]
pub struct FriendlyRecord<T> {
    ttl: i64,
    data: T,
}

impl<T: Serialize + HasRType> Serialize for FriendlyRecord<T> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut state = serializer.serialize_struct("FriendlyRecord", 3)?;
        state.serialize_field("ttl", &self.ttl)?;
        state.serialize_field("data", &self.data)?;
        state.serialize_field("rtype", &self.data.rtype())?;
        state.end()
    }
}

impl<T> FriendlyRecord<T> {
    fn new(ttl: u32, data: T) -> Self {
        FriendlyRecord {
            ttl: ttl.into(),
            data,
        }
    }
}

#[derive(Debug)]
pub struct FriendlyRecordSet<T> {
    ttl: i64,
    data: Vec<T>,
}

impl<T: Serialize + HasRType> Serialize for FriendlyRecordSet<T> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut state = serializer.serialize_struct("FriendlyRecordSet", 3)?;
        state.serialize_field("ttl", &self.ttl)?;
        state.serialize_field("data", &self.data)?;
        state.serialize_field("rtype", &self.data.first().map(|d| d.rtype().to_string()))?;
        state.end()
    }
}

impl<T> FriendlyRecordSet<T> {
    fn new(ttl: u32) -> Self {
        FriendlyRecordSet {
            ttl: ttl.into(),
            data: Vec::new()
        }
    }
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(rename_all="lowercase")]
pub enum ConfigurationType {
    Mail,
    Web,
}

#[derive(Debug, Serialize)]
pub struct MailConfiguration {
    pub servers: Option<FriendlyRecordSet<MailServer>>,
    pub spf: Option<FriendlyRecord<Spf>>,
}

impl MailConfiguration {
    pub fn new() -> Self {
        MailConfiguration {
            servers: None,
            spf: None
        }
    }
}

#[derive(Debug, Serialize)]
pub struct WebConfiguration {
    pub addresses: Option<FriendlyRecordSet<Address>>,
}


impl WebConfiguration {
    pub fn new() -> Self {
        WebConfiguration {
            addresses: None,
        }
    }
}

#[derive(Debug, Serialize)]
pub struct FriendlyRecordSetGroup {
    owner: String,
    general_records: Vec<FriendlyRecordSet<FriendlyRData>>,
    #[serde(serialize_with = "as_vector")]
    services: HashMap<ServiceType, FriendlyRecordSet<Service>>,
    mail: Option<MailConfiguration>,
    web: Option<WebConfiguration>,
}

fn as_vector<S>(services: &HashMap<ServiceType, FriendlyRecordSet<Service>>, ser: S) -> Result<S::Ok, S::Error>
    where S: Serializer
{
    let container: Vec<_> = services.iter().collect();
    serde::Serialize::serialize(&container, ser)
}

impl FriendlyRecordSetGroup {
    pub fn new(owner: String) -> Self {
        FriendlyRecordSetGroup {
            owner,
            general_records: Vec::new(),
            services: HashMap::new(),
            mail: None,
            web: None,
        }
    }
}

#[derive(Debug, Serialize)]
pub struct FriendlyRecords {
    records: Vec<FriendlyRecordSetGroup>,
    aliases: Vec<Alias>,
}

impl FriendlyRecords {
    pub fn new() -> Self {
        FriendlyRecords {
            records: Vec::new(),
            aliases: Vec::new(),
        }
    }
}

impl From<internal::RecordList> for FriendlyRecords {
    fn from(value: internal::RecordList) -> Self {

        let mut records = FriendlyRecords::new();

        for record in value.records {
            let internal::Record { name, ttl, rdata } = record;
            let name = name.to_string();
            let rdata = rdata.friendly(&name, ttl);

            if rdata.is_none() {
                continue;
            }

            let (name, rdata) = rdata.unwrap();

            if let FriendlyRData::Alias(alias) = rdata {
                records.aliases.push(alias)
            } else {

                if records.records.is_empty() {
                    records.records.push(FriendlyRecordSetGroup::new(name));
                } else {
                    let group = records.records.last().unwrap();

                    if group.owner != name {
                        records.records.push(FriendlyRecordSetGroup::new(name));
                    }
                }

                let current_group = records.records.last_mut().unwrap();

                match rdata {
                    FriendlyRData::Address(address) => {
                        let web = current_group.web.get_or_insert_with(WebConfiguration::new);
                        let addresses = web.addresses.get_or_insert_with(|| FriendlyRecordSet::new(ttl));

                        addresses.data.push(address);
                    },
                    FriendlyRData::MailServer(mailserver) => {
                        let mail: &mut MailConfiguration = current_group.mail.get_or_insert_with(MailConfiguration::new);
                        let servers = mail.servers.get_or_insert_with(|| FriendlyRecordSet::new(ttl));

                        servers.data.push(mailserver);
                    },
                    FriendlyRData::Spf(spf) => {
                        let mail: &mut MailConfiguration = current_group.mail.get_or_insert_with(MailConfiguration::new);

                        mail.spf = Some(FriendlyRecord::new(ttl, spf));
                    },
                    FriendlyRData::Service(service) => {
                        let services = current_group.services.entry(service.service_type.clone()).or_insert(FriendlyRecordSet::new(ttl));

                        services.data.push(service)
                    }
                    FriendlyRData::Alias(_) => {},
                    data => {
                        // TODO: NS -> Skip if NS for zone (authority), create Delegation section with glue + DS for others (how to check if record is glue?)

                        if current_group.general_records.is_empty() {
                            current_group.general_records.push(FriendlyRecordSet::new(ttl));
                        } else {
                            let rrset = current_group.general_records.last().unwrap();
                            if rrset.data.last().unwrap().rtype() != data.rtype() {
                                current_group.general_records.push(FriendlyRecordSet::new(ttl));
                            }
                        }

                        let current_rrset = current_group.general_records.last_mut().unwrap();
                        current_rrset.data.push(data)
                    }
                }
            }
        }

        records
    }
}

/* --------- RDATA --------- */

#[derive(Debug, Deserialize, Serialize)]
pub struct Address {
    pub address: String
}


impl HasRType for Address {
    fn rtype(&self) -> FriendlyRType {
        FriendlyRType::Address
    }
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Service {
    pub service_type: ServiceType,
    pub port: i64,
    pub weight: i64,
    pub priority: i64,
    pub server: String,
}

#[derive(Debug, Deserialize, Serialize, Hash, Eq, PartialEq, Clone)]
#[serde(rename_all = "lowercase", tag = "service_type")]
pub enum ServiceType {
    Other { protocol: String, name: String },
}

impl HasRType for Service {
    fn rtype(&self) -> FriendlyRType {
        FriendlyRType::Service
    }
}

#[derive(Debug, Deserialize, Serialize)]
pub struct MailServer {
    pub preference: i64,
    pub mail_exchanger: String,
}

impl HasRType for MailServer {
    fn rtype(&self) -> FriendlyRType {
        FriendlyRType::MailServer
    }
}

#[derive(Debug, Deserialize, Serialize)]
pub struct NameServer {
    pub target: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct TextData {
    pub text: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Alias {
    pub from: String,
    pub target: String,
    pub ttl: i64,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Spf {
    pub policy: String
}

impl HasRType for Spf {
    fn rtype(&self) -> FriendlyRType {
        FriendlyRType::Spf
    }
}
