use async_trait::async_trait;
use serde::{Deserialize, Serialize};
use rusqlite::Error as RusqliteError;


use crate::database::{BoxedDb, sqlite::SqliteDB};
use crate::dns::{BoxedZoneDriver, BoxedRecordDriver, DnsDriverError};
use crate::errors::Error;
use crate::macros::push_error;
use crate::resources::dns::internal::RecordList;
use crate::validation;

pub enum ZoneError {
    ZoneConflict { name: String },
    NotFound { name: String },
    NotExistsNs { name: String },
    Validation { suberrors: Vec<Error> },
}

#[derive(Debug, Serialize)]
pub struct Zone {
    pub name: String,
}

impl Zone {
    pub async fn create(create_zone: CreateZoneRequest, zone_driver: BoxedZoneDriver, db: BoxedDb) -> Result<Self, Error> {
        let create_zone = create_zone.validate()?;

        zone_driver.zone_exists(&create_zone.name)
            .await
            .map_err(|e| {
                match e {
                    DnsDriverError::ZoneNotFound { name } => {
                        Error::from(ZoneError::NotExistsNs { name })
                            .with_path("/name")
                    },
                    e => Error::from(e)
                }
            })?;

        db.create_zone(create_zone).await
    }

    pub async fn get_records(&self, record_driver: BoxedRecordDriver) ->Result<RecordList, Error> {
        let mut records = record_driver.get_records(&self.name).await?;

        records.sort();

        Ok(records)
    }
}

#[derive(Deserialize)]
pub struct CreateZoneRequest {
    pub name: String
}

pub struct CreateZone {
    pub name: String
}

impl CreateZoneRequest {
    pub fn validate(self) -> Result<CreateZone, Error> {

        let mut errors = Vec::new();

        let name = push_error!(validation::normalize_domain(&self.name), errors, "/name");
        name.ok_or(Error::from(ZoneError::Validation { suberrors: errors }))
            .map(|name| {
                CreateZone { name }
            })
    }
}

#[async_trait]
pub trait ZoneModel: Send + Sync {
    async fn create_zone(&self, create_zone: CreateZone) -> Result<Zone, Error>;
    async fn get_zone_by_name(&self, zone_name: &str) -> Result<Zone, Error>;
}

#[async_trait]
impl ZoneModel for SqliteDB {
    async fn create_zone(&self, create_zone: CreateZone) -> Result<Zone, Error> {
        let pool = self.pool.clone();

        let conn = pool.get().await?;

        tokio::task::block_in_place(move || {
            let mut stmt = conn.prepare("insert into zones (name) values (?1) returning *")?;
            let zone = stmt.query_row((&create_zone.name,), |row| {
                Ok(Zone {
                    name: row.get(0)?
                })
            }).map_err(|e| {
                match e {
                    /* SQLITE_CONSTRAINT_PRIMARYKEY */
                    RusqliteError::SqliteFailure(e, _) if e.extended_code == 1555 => {
                        Error::from(ZoneError::ZoneConflict { name: create_zone.name })
                            .with_path("/name")
                    },
                    e => Error::new("internal:zone:create", "Failed to create zone")
                        .with_cause(&e.to_string())
                }
            })?;

            Ok(zone)
        })
    }

    async fn get_zone_by_name(&self, zone_name: &str) -> Result<Zone, Error> {
        let pool = self.pool.clone();

        let conn = pool.get().await?;

        tokio::task::block_in_place(move || {
            let mut stmt = conn.prepare("select * from zones where name = ?1")?;
            let zone = stmt.query_row((zone_name,), |row| {
                Ok(Zone {
                    name: row.get(0)?
                })
            }).map_err(|e| {
                match e {
                    RusqliteError::QueryReturnedNoRows => {
                        Error::from(ZoneError::NotFound { name: zone_name.to_string() })
                    },
                    e => Error::new("internal:zone:get_by_name", "Failed to fetch zone by name")
                        .with_cause(&e.to_string())
                }
            })?;

            Ok(zone)
        })
    }
}


/*
use crate::models::user::UserInfo;

use uuid::Uuid;
use diesel::prelude::*;
use diesel::result::Error as DieselError;
use serde::{Serialize, Deserialize};

use crate::schema::*;
use super::name::AbsoluteName;
use super::user::UserZone;
use super::errors::UserError;


#[derive(Debug, Serialize, Queryable, Identifiable, Insertable)]
#[table_name = "zone"]
pub struct Zone {
    #[serde(skip)]
    pub id: String,
    pub name: String,
}

#[derive(Debug, Deserialize)]
pub struct AddZoneMemberRequest {
    pub id: String,
}

#[derive(Debug, Deserialize, FromForm)]
pub struct CreateZoneRequest {
    pub name: AbsoluteName,
}

// NOTE: Should probably not be implemented here
//       also, "UserError" seems like a misleading name
impl Zone {
    pub fn get_all(conn: &diesel::SqliteConnection) -> Result<Vec<Zone>, UserError> {
        use crate::schema::zone::dsl::*;

        zone.get_results(conn)
            .map_err(UserError::DbError)
    }

    pub fn get_by_name(conn: &diesel::SqliteConnection, zone_name: &str) -> Result<Zone, UserError> {
        use crate::schema::zone::dsl::*;

        zone.filter(name.eq(zone_name))
            .get_result(conn)
            .map_err(|e| match e {
                DieselError::NotFound => UserError::ZoneNotFound,
                other => UserError::DbError(other)
            })
    }

    pub fn create_zone(conn: &diesel::SqliteConnection, zone_request: CreateZoneRequest) -> Result<Zone, UserError> {
        use crate::schema::zone::dsl::*;

        let new_zone = Zone {
            id: Uuid::new_v4().to_simple().to_string(),
            name: zone_request.name.to_utf8(),
        };

        diesel::insert_into(zone)
            .values(&new_zone)
            .execute(conn)
            .map_err(|e| match e {
                DieselError::DatabaseError(diesel::result::DatabaseErrorKind::UniqueViolation, _) => UserError::UserConflict,
                other => UserError::DbError(other)
            })?;
        Ok(new_zone)
    }


    pub fn add_member(&self, conn: &diesel::SqliteConnection, new_member: &UserInfo) -> Result<(), UserError> {
        use crate::schema::user_zone::dsl::*;

        let new_user_zone = UserZone {
            zone_id: self.id.clone(),
            user_id: new_member.id.clone()
        };

        let res = diesel::insert_into(user_zone)
            .values(new_user_zone)
            .execute(conn);

        match res {
            // If user has already access to the zone, safely ignore the conflit
            // TODO: use 'on conflict do nothing' in postgres when we get there
            Err(DieselError::DatabaseError(diesel::result::DatabaseErrorKind::UniqueViolation, _)) => (),
            Err(e) => return Err(e.into()),
            Ok(_) => ()
        };
        Ok(())
    }
}
*/
