use crate::errors::Error;

pub enum DomainValidationError {
    EmptyDomain,
    DomainTooLong { length: usize },
    CharactersNotPermitted { label: String },
    LabelToolLong { length: usize, label: String },
    EmptyLabel
}


/// Not complete but probably good enough
/// https://doc.zonemaster.fr/v2024.1/specifications/tests/RequirementsAndNormalizationOfDomainNames.html
/// TODO: No support of dots in labels, how to handle RNAME in SOA?
pub fn normalize_domain(domain_name: &str) -> Result<String, Error> {
    if domain_name.is_empty() {
        return Err(Error::from(DomainValidationError::EmptyDomain))
    }

    let domain = domain_name.strip_suffix('.').unwrap_or(domain_name).to_lowercase();

    if domain.len() > 255 {
        Err(Error::from(DomainValidationError::DomainTooLong { length: domain.len() }))
    } else {
        let labels = domain.split('.').collect::<Vec<_>>();

        if labels.iter().any(|l| l.is_empty()) {
            return Err(
                Error::from(DomainValidationError::EmptyLabel)
            );
        }

        for label in labels {
            if !label.chars().all(|c| {
                // allow for '/' for reverse zone
                c.is_ascii_alphanumeric() || c == '-' || c == '/' || c == '_'
            }) {
                return Err(
                    Error::from(DomainValidationError::CharactersNotPermitted { label: label.into() })
                );
            }

            if label.len() > 63 {
                return Err(Error::from(DomainValidationError::LabelToolLong {
                    label: label.into(),
                    length: label.len()
                }));
            }
        }

        Ok(domain)
    }

}
pub enum TxtParseError {
    MissingEscape,
    NonAscii { character: String },
    BadEscapeDigitsTooShort { sequence: String },
    BadEscapeDigitsNotDigits { sequence: String },
    BadEscapeDigitIndexTooHigh { sequence: String },
}

pub fn parse_txt_data(text: &str) -> Result<Vec<u8>, Vec<Error>> {
    let mut chars = text.chars();
    let mut errors = Vec::new();
    let mut data = Vec::new();

    #[inline]
    fn printable(ch: char) -> bool {
        ch.is_ascii() && ('\u{20}'..='\u{7E}').contains(&ch)
    }

    while let Some(ch) = chars.next() {
        if ch == '\\' {
            match chars.next() {
                Some(ch) => {
                    if ch.is_ascii_digit() {
                        let mut digits: Vec<_> = chars.by_ref().take(2).collect();
                        digits.insert(0, ch);
                        if digits.len() < 3 {
                            errors.push(Error::from(TxtParseError::BadEscapeDigitsTooShort { sequence: String::from_iter(digits) }))
                        } else if digits.iter().all(|c| c.is_ascii_digit()) {
                            errors.push(Error::from(TxtParseError::BadEscapeDigitsNotDigits { sequence: String::from_iter(digits) }))
                        } else {
                            let index = {
                                digits[0].to_digit(10).unwrap() * 100 +
                                digits[1].to_digit(10).unwrap() * 10 +
                                digits[2].to_digit(10).unwrap()
                            };

                            if index > 255 {
                                errors.push(Error::from(TxtParseError::BadEscapeDigitIndexTooHigh { sequence: String::from_iter(digits) }))
                            }
                        }
                    } else if printable(ch) {
                        data.push(ch as u8)
                    } else {
                        errors.push(Error::from(TxtParseError::NonAscii { character: ch.into() }))
                    }
                },
                None => {
                    errors.push(Error::from(TxtParseError::MissingEscape))
                }
            }
        } else if printable(ch) {
            data.push(ch as u8);
        } else {
            errors.push(Error::from(TxtParseError::NonAscii { character: ch.into() }))
        }
    }

    //TODO: check txt data max length?

    if errors.is_empty() {
        Ok(data)
    } else {
        Err(errors)
    }
}
